@@ -5,14 +5,15 @@ Author: Nikolay Lysenko
 """
 
 
+import itertools
 import math
 import random
 from dataclasses import dataclass
-from typing import Optional
+from typing import Any, Optional
 
 from sinethesizer.utils.music_theory import get_note_to_position_mapping
 
-from .utils import (
+from .music_theory import (
     N_SEMITONES_PER_OCTAVE,
     TONE_ROW_LEN,
     get_smallest_intervals_between_pitch_classes,
@@ -61,12 +62,14 @@ class FragmentParams:
     upper_line_highest_note: str
     upper_line_lowest_note: str
     pauses_fraction: float
+    temporal_content: Optional[dict[int, dict[str, Any]]] = None
+    sonic_content: Optional[dict[int, dict[str, Any]]] = None
 
 
 @dataclass
 class Fragment:
     """A fragment of a musical piece."""
-    temporal_content: list[list[list[Event]]]
+    temporal_content: list[list[float]]
     sonic_content: list[list[str]]
     meter_numerator: int
     meter_denominator: int
@@ -74,14 +77,49 @@ class Fragment:
     line_ids: list[int]
     upper_line_highest_position: int
     upper_line_lowest_position: int
+    n_melodic_lines_by_group: list[int]
     n_tone_row_instances_by_group: list[int]
+    mutable_temporal_content_indices: list[int]
+    mutable_sonic_content_indices: list[int]
     melodic_lines: Optional[list[list[Event]]] = None
     sonorities: Optional[list[list[Event]]] = None
 
+    def __eq__(self, other: Any):
+        if not isinstance(other, Fragment):
+            return NotImplemented
+        return (
+            self.temporal_content == other.temporal_content
+            and self.sonic_content == other.sonic_content
+        )
+
+
+def validate_initialized_content(params: FragmentParams) -> None:
+    """
+    Validate parameters related to initialized parts of temporal content and sonic content.
+
+    :param params:
+        parameters of a fragment to be created
+    :return:
+        None
+    """
+    if params.temporal_content is not None:
+        total_duration_in_beats = params.n_measures * params.meter_numerator
+        for line_index, line_params in params.temporal_content.items():
+            duration_in_beats = sum(line_params['durations'])
+            if duration_in_beats != total_duration_in_beats:
+                raise ValueError("A line has duration that is not equal to that of the fragment.")
+    if params.sonic_content is not None:  # pragma: no branch
+        for group_index, group_params in params.sonic_content.items():
+            n_sound_events = len([x for x in group_params['pitch_classes'] if x != 'pause'])
+            n_tone_row_instances = n_sound_events / TONE_ROW_LEN
+            declared_n_instances = params.groups[group_index]['n_tone_row_instances']
+            if n_tone_row_instances != declared_n_instances:
+                raise ValueError("A group has wrong number of tone row instances.")
+
 
 def validate(params: FragmentParams) -> None:
     """
-    Validate parameters.
+    Validate all parameters.
 
     :param params:
         parameters of a fragment to be created
@@ -95,6 +133,7 @@ def validate(params: FragmentParams) -> None:
         raise ValueError("IDs of melodic lines must be unique.")
     if float(params.meter_numerator) not in SUPPORTED_DURATIONS:
         raise ValueError(f"Meter numerator = {params.meter_numerator} is not supported.")
+    validate_initialized_content(params)
 
 
 def split_time_span(n_measures: int, n_events: int, meter_numerator: float) -> list[float]:
@@ -127,59 +166,93 @@ def split_time_span(n_measures: int, n_events: int, meter_numerator: float) -> l
     return durations
 
 
-def find_initial_durations(params: FragmentParams) -> list[list[list[float]]]:
+def calculate_number_of_undefined_events(
+        group_index: int,
+        temporal_content: list[list[float]],
+        sonic_content: dict[int, dict[str, Any]],
+        line_indices: list[int],
+        n_tone_row_instances: int,
+        pauses_fraction: float
+) -> int:
+    """
+    Calculate total number of undefined events in a group of melodic lines sharing the same series.
+
+    :param group_index:
+        index of a group of melodic lines
+    :param temporal_content:
+        partially filled list of lists of event durations (in reference beats)
+        for each melodic line
+    :param sonic_content:
+        mapping from group index to user-defined sequence of its pitch classes (including pauses)
+    :param line_indices:
+        indices of melodic lines forming the group
+    :param n_tone_row_instances:
+        number of tone row instances in the group
+    :param pauses_fraction:
+        fraction of pauses amongst all events from the group
+    :return:
+        total number of events in the group that are not initially defined by a user
+    """
+    if group_index in sonic_content:
+        return len(sonic_content[group_index]['pitch_classes'])
+    n_sound_events = n_tone_row_instances * TONE_ROW_LEN
+    n_events = int(round(n_sound_events / (1 - pauses_fraction)))
+    n_events -= sum(len(temporal_content[line_index]) for line_index in line_indices)
+    return n_events
+
+
+def distribute_events_between_lines(n_events: int, n_lines: int) -> list[int]:
+    """
+    Distribute evenly the specified number of events between the specified number of lines.
+
+    :param n_events:
+        number of events
+    :param n_lines:
+        number of lines
+    :return:
+        list of numbers of events in a particular line
+    """
+    results = [n_events // n_lines for _ in range(n_lines)]
+    i = 0
+    while sum(results) < n_events:
+        results[i] += 1
+        i += 1
+    return results
+
+
+def create_initial_temporal_content(params: FragmentParams) -> list[list[float]]:
     """
     Split time span of each melodic line into durations of individual events.
 
     :param params:
         parameters of a fragment to be created
     :return:
-        lists of event durations (in reference beats) for each melodic line from each group of
-        melodic lines sharing the same series (in terms of vertical distribution of series pitches)
+        lists of event durations (in reference beats) for each melodic line
     """
+    temporal_content = [
+        params.temporal_content.get(line_index, {}).get('durations', [])
+        for line_index, line_id in enumerate(params.line_ids)
+    ]
     meter_numerator = float(params.meter_numerator)
-    results = []
-    for group_params in params.groups:
-        nested_results = []
-        n_sound_events = group_params['n_tone_row_instances'] * TONE_ROW_LEN
-        n_events = int(round(n_sound_events / (1 - params.pauses_fraction)))
+    group_lines_start_index = 0
+    for group_index, group_params in enumerate(params.groups):
         n_lines = group_params['n_melodic_lines']
-        n_events_per_line = [n_events // n_lines for _ in range(n_lines)]
-        line_index = 0
-        while sum(n_events_per_line) < n_events:
-            n_events_per_line[line_index] += 1
-            line_index += 1
-        for current_n_events in n_events_per_line:
-            durations = split_time_span(params.n_measures, current_n_events, meter_numerator)
-            nested_results.append(durations)
-        results.append(nested_results)
-    return results
-
-
-def find_initial_temporal_content(params: FragmentParams) -> list[list[list[Event]]]:
-    """
-    Find initial value of data structure that keeps track of event durations.
-
-    :param params:
-        parameters of a fragment to be created
-    :return:
-        list where for each group of melodic lines sharing the same series for each melodic line
-        there are events representing temporal (loosely speaking, rhythmic) structure of the line
-    """
-    line_index = 0
-    temporal_content = []
-    groups_durations = find_initial_durations(params)
-    for group_durations in groups_durations:
-        group_content = []
-        for line_durations in group_durations:
-            start_time = 0
-            line_content = []
-            for duration in line_durations:
-                line_content.append(Event(line_index, start_time, duration))
-                start_time += duration
-            group_content.append(line_content)
-            line_index += 1
-        temporal_content.append(group_content)
+        line_indices = list(range(group_lines_start_index, group_lines_start_index + n_lines))
+        n_undefined_events = calculate_number_of_undefined_events(
+            group_index,
+            temporal_content,
+            params.sonic_content,
+            line_indices,
+            group_params['n_tone_row_instances'],
+            params.pauses_fraction
+        )
+        indices_of_undefined_lines = [x for x in line_indices if x not in params.temporal_content]
+        n_undefined_lines = len(indices_of_undefined_lines)
+        n_events_per_line = distribute_events_between_lines(n_undefined_events, n_undefined_lines)
+        for line_index, n_events in zip(indices_of_undefined_lines, n_events_per_line):
+            durations = split_time_span(params.n_measures, n_events, meter_numerator)
+            temporal_content[line_index] = durations
+        group_lines_start_index += n_lines
     return temporal_content
 
 
@@ -206,33 +279,78 @@ def replicate_tone_row(tone_row: list[str], n_instances: int) -> list[str]:
     return pitch_classes
 
 
-def find_initial_sonic_content(params: FragmentParams) -> list[list[str]]:
+def create_initial_sonic_content(
+        params: FragmentParams, temporal_content: list[list[float]]
+) -> list[list[str]]:
     """
-    Find initial value of data structure that keeps track of pitch classes.
+    Create initial data structure that keeps track of pitch classes.
 
     :param params:
         parameters of a fragment to be created
+    :param temporal_content:
+        lists of event durations (in reference beats) for each melodic line
     :return:
         list where for each group of melodic lines sharing the same series sonic content
         of the series is stored (i.e., there is a sequence consisting of pitch classes and pauses)
     """
     sonic_content = []
-    for group_params in params.groups:
-        group_content = []
+    group_lines_start_index = 0
+    for group_index, group_params in enumerate(params.groups):
+        n_lines = group_params['n_melodic_lines']
+        if group_index in params.sonic_content:
+            sonic_content.append(params.sonic_content[group_index]['pitch_classes'])
+            group_lines_start_index += n_lines
+            continue
+
+        n_events = 0
+        durations = temporal_content[group_lines_start_index:(group_lines_start_index + n_lines)]
+        for line_durations in durations:
+            n_events += len(line_durations)
         n_sound_events = group_params['n_tone_row_instances'] * TONE_ROW_LEN
-        n_events = int(round(n_sound_events / (1 - params.pauses_fraction)))
         n_pauses = n_events - n_sound_events
+        if n_pauses < 0:
+            raise ValueError("User-defined temporal content has not enough events.")
+
         pauses_indices = random.sample(range(n_events), n_pauses)
         series = replicate_tone_row(params.tone_row, group_params['n_tone_row_instances'])
+        group_sonic_content = []
         for index in range(n_events):
-            if index in pauses_indices:
-                group_content.append('pause')
-            else:
-                group_content.append(series.pop(0))
-        sonic_content.append(group_content)
+            group_sonic_content.append('pause' if index in pauses_indices else series.pop(0))
+        sonic_content.append(group_sonic_content)
+
+        group_lines_start_index += n_lines
     return sonic_content
 
 
+def create_grouped_rhythm_only_lines(fragment: Fragment) -> list[list[list[Event]]]:
+    """
+    Create rhythm-only lines grouped by sharing of the same series.
+
+    :param fragment:
+        fragment with non-empty temporal content
+    :return:
+        list where for each group of melodic lines sharing the same series for each melodic line
+        there are events representing temporal (loosely speaking, rhythmic) structure of the line,
+        but keeping no information about pitches
+    """
+    line_index = 0
+    rhythm_only_lines = []
+    end_indices = list(itertools.accumulate(fragment.n_melodic_lines_by_group))
+    start_indices = [0] + end_indices[:-1]
+    for start_index, end_index in zip(start_indices, end_indices):
+        group_of_lines = []
+        for line_durations in fragment.temporal_content[start_index:end_index]:
+            start_time = 0
+            line = []
+            for duration in line_durations:
+                line.append(Event(line_index, start_time, duration))
+                start_time += duration
+            group_of_lines.append(line)
+            line_index += 1
+        rhythm_only_lines.append(group_of_lines)
+    return rhythm_only_lines
+
+
 def distribute_pitch_classes(fragment: Fragment) -> list[list[Event]]:
     """
     Distribute pitch classes across melodic lines.
@@ -244,12 +362,13 @@ def distribute_pitch_classes(fragment: Fragment) -> list[list[Event]]:
         these lines are derived from temporal content and sonic content of the fragment
     """
     melodic_lines = [[] for _ in fragment.line_ids]
-    zipped = zip(fragment.temporal_content, fragment.sonic_content)
-    for group_temporal_content, group_sonic_content in zipped:
+    grouped_rhythm_only_lines = create_grouped_rhythm_only_lines(fragment)
+    zipped = zip(grouped_rhythm_only_lines, fragment.sonic_content)
+    for group_of_rhythm_only_lines, group_sonic_content in zipped:
         timeline = [
             event
-            for line_temporal_content in group_temporal_content
-            for event in line_temporal_content
+            for rhythm_only_line in group_of_rhythm_only_lines
+            for event in rhythm_only_line
         ]
         timeline = sorted(timeline, key=lambda event: (event.start_time, event.line_index))
         for event, pitch_class in zip(timeline, group_sonic_content):
@@ -415,6 +534,38 @@ def override_calculated_attributes(fragment: Fragment) -> Fragment:
     return fragment
 
 
+def find_mutable_temporal_content_indices(params: FragmentParams) -> list[int]:
+    """
+    Find indices of melodic lines such that their temporal content can be modified.
+
+    :param params:
+        parameters of a fragment to be created
+    :return:
+        indices of melodic lines such that their temporal content can be modified
+    """
+    results = []
+    for line_index, line_id in enumerate(params.line_ids):
+        if not params.temporal_content.get(line_index, {}).get('immutable', False):
+            results.append(line_index)
+    return results
+
+
+def find_mutable_sonic_content_indices(params: FragmentParams) -> list[int]:
+    """
+    Find indices of groups such that their sonic content can be modified.
+
+    :param params:
+        parameters of a fragment to be created
+    :return:
+        indices of groups such that their sonic content can be modified
+    """
+    results = []
+    for group_index, group_params in enumerate(params.groups):
+        if not params.sonic_content.get(group_index, {}).get('immutable', False):
+            results.append(group_index)
+    return results
+
+
 def initialize_fragment(params: FragmentParams) -> Fragment:
     """
     Create fragment by its parameters.
@@ -424,17 +575,24 @@ def initialize_fragment(params: FragmentParams) -> Fragment:
     :return:
         fragment
     """
+    params.temporal_content = params.temporal_content or {}
+    params.sonic_content = params.sonic_content or {}
     validate(params)
+    temporal_content = create_initial_temporal_content(params)
+    sonic_content = create_initial_sonic_content(params, temporal_content)
     fragment = Fragment(
-        find_initial_temporal_content(params),
-        find_initial_sonic_content(params),
+        temporal_content,
+        sonic_content,
         params.meter_numerator,
         params.meter_denominator,
         params.n_measures * params.meter_numerator,
         params.line_ids,
         NOTE_TO_POSITION_MAPPING[params.upper_line_highest_note],
         NOTE_TO_POSITION_MAPPING[params.upper_line_lowest_note],
-        [group['n_tone_row_instances'] for group in params.groups]
+        [group['n_melodic_lines'] for group in params.groups],
+        [group['n_tone_row_instances'] for group in params.groups],
+        find_mutable_temporal_content_indices(params),
+        find_mutable_sonic_content_indices(params),
     )
     fragment = override_calculated_attributes(fragment)
     return fragment
